# Python imports
import os
import signal
import json
import subprocess
import threading

# Lib imports
import gi
gi.require_version('Gtk', '3.0')
gi.require_version('GtkSource', '4')
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import GtkSource

# Application imports
# from libs.dto.lsp_structs import TextDocumentItem
from libs.dto.lsp_message_structs import MessageEncoder, LSPRequest, LSPNotification, LSPResponse
from .buttons.top_button_box import TopButtonBox
from .enteries.lsp_message_source_view import LspMessageSourceView
from .buttons.bottom_button_box import BottomButtonBox



LEN_HEADER  = "Content-Length: "
TYPE_HEADER = "Content-Type: "



class LSPMessageBox(Gtk.Box):
    def __init__(self):
        super(LSPMessageBox, self).__init__()

        self._lsp_pid      = -1
        self._message_id   = 0

        # https://github.com/microsoft/multilspy/tree/main/src/multilspy/language_servers
        # initialize-params-slim.json was created off of jedi_language_server one
        self._lsp_init_ops = settings_manager.get_lsp_init_data()

        self.read_lock     = threading.Lock()
        self.write_lock    = threading.Lock()

        self._setup_styling()
        self._setup_signals()
        self._subscribe_to_events()
        self._load_widgets()

        self.show()


    def _setup_styling(self):
        ctx = self.get_style_context()
        ctx.add_class("lsp-message-box")

        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.set_hexpand(True)
        self.set_vexpand(True)

    def _setup_signals(self):
        ...

    def _subscribe_to_events(self):
        ...

    def _load_widgets(self):
        scrolled_win        = Gtk.ScrolledWindow()
        self.lsp_msg_src_vw = LspMessageSourceView()
        self.top_buttons    = TopButtonBox()
        self.bottom_buttons = BottomButtonBox()

        scrolled_win.show_all()

        scrolled_win.add(self.lsp_msg_src_vw)
        self.add(self.top_buttons)
        self.add(scrolled_win)
        self.add(self.bottom_buttons)

    def update_message_id_label(self):
        self.top_buttons.update_message_id_lbl(self._message_id)

    def send_initialize_message(self):
        parent            = self.get_parent()
        folder_choser_btn = self.top_buttons.folder_choser_btn

        init_ops          = parent.init_ops_src_vw.get_text_str()
        workspace_file    = folder_choser_btn.get_file().get_path()
        workspace_uri     = folder_choser_btn.get_uri()
        folder_name       = os.path.basename(workspace_file)

        self._lsp_init_ops["processId"]        = settings_manager.get_app_pid()
        self._lsp_init_ops["rootPath"]         = workspace_file
        self._lsp_init_ops["rootUri"]          = workspace_uri
        self._lsp_init_ops["workspaceFolders"] = [
            {
                "name": folder_name,
                "uri": workspace_uri
            }
        ]

        self._lsp_init_ops["initializationOptions"] = json.loads(init_ops)
        self.send_request("initialize", self._lsp_init_ops)

        del init_ops
        del workspace_file
        del workspace_uri
        del folder_name
        del parent

    def send_initialized_message(self):
        self.send_notification("initialized")

    def send_notification(self, method: str, params: {} = {}):
        self._send_message( LSPNotification(method, params) )

    def send_request(self, method: str, params: {} = {}):
        self._monitor_lsp_response()

        self._message_id += 1
        self._send_message( LSPRequest(self._message_id, method, params) )

    def _send_message(self, data: LSPRequest or LSPNotification):
        if not data or not hasattr(self, "lsp_process"): return

        message_str  = json.dumps(data, cls = MessageEncoder)
        message_size = len(message_str)
        message      = f"Content-Length: {message_size}\r\n\r\n{message_str}"

        self.get_parent().log_list.add_log_entry("Client", data)

        with self.write_lock:
            self.lsp_process.stdin.write( message.encode("utf-8") )
            self.lsp_process.stdin.flush()

        self.update_message_id_label()

        del message_str
        del message_size
        del message

    def button_send_notification(self):
        message = json.loads( self.lsp_msg_src_vw.get_text_str() )
        self.send_notification(message["method"], message["params"])

    def button_send_request(self):
        message = json.loads( self.lsp_msg_src_vw.get_text_str() )
        self.send_request(message["method"], message["params"])

    def start_stop_lsp(self):
        if self._lsp_pid == -1:
            pid = self.start_lsp()
            if not pid: return

            self._lsp_pid = pid
            self._monitor_lsp_response()
        else:
            self.stop_lsp()
            self.get_parent().log_list.clear()

    def start_lsp(self):
        parent           = self.get_parent()
        _command: str    = parent.alt_command_entry.get_text()
        # _command: str    = parent.command_entry.get_text()
        # _command: str    = parent.socket_entry.get_text()
        command: []      = _command.split() if len( _command.split() ) > 0 else [ _command ]
        try:
            self.lsp_process = subprocess.Popen(
                command,
                stdout = subprocess.PIPE,
                stdin  = subprocess.PIPE
            )
        except Exception as e:
            self.bottom_buttons.rest_buttons()
            self.get_parent().log_list.add_log_entry(
                "LSP Client Error",
                LSPResponse(
                    None,
                    {
                        "error": repr(e)
                    }
                )
            )
            return

        return self.lsp_process.pid

    def stop_lsp(self):
        if self._lsp_pid == -1: return

        self._lsp_pid    = -1
        self._message_id = 0
        self.lsp_process.terminate()
        self.update_message_id_label()

    # https://github.com/sr-lab/coqpyt/blob/master/coqpyt/lsp/json_rpc_endpoint.py#L65
    # Virtually this whole method unabashedly taken from ^ ...
    @daemon_threaded
    def _monitor_lsp_response(self):
        if not hasattr(self, "lsp_process"): return

        with self.read_lock:
            message_size = None
            while True:
                line = self.lsp_process.stdout.readline()
                if not line: return None  # Quit listener... 

                line = line.decode("utf-8")
                if not line.endswith("\r\n"):
                    raise Exception(
                        "Bad header: missing newline"
                    )
                line = line[:-2]  # Strip the "\r\n"
                if line == "":    # We're done with the headers...
                    break
                elif line.startswith(LEN_HEADER):
                    line = line[len(LEN_HEADER) :]
                    if not line.isdigit():
                        raise Exception(
                            "Bad header: size is not int",
                        )
                    message_size = int(line)
                elif line.startswith(TYPE_HEADER):
                    # Not doing anything with type header, currently...
                    pass
                else:
                    line = line.split(LEN_HEADER)
                    if len(line) == 2:
                        message_size = line[1]
                    raise Exception(
                        "Bad header: unknown header"
                    )

            if not message_size: return

            data        = self.lsp_process.stdout.read(message_size)
            jsonrpc_res = data.decode("utf-8")
            lsp_response  = json.loads( jsonrpc_res )
            response_id = -1

            if not lsp_response: return
            if "id" in lsp_response.keys(): response_id = lsp_response["id"]

            GLib.idle_add(self.handle_lsp_response, LSPResponse(response_id, lsp_response))

    def handle_lsp_response(self, lsp_response: LSPResponse):
        self.get_parent().log_list.add_log_entry("LSP Response", lsp_response)
        result = lsp_response.result
        keys   = result.keys()

        if "error" in keys:
            lsp_response = result["error"]
            logger.debug(f"LSP Error Code:  {result['code']}")
            logger.debug(f"LSP Error Message:\n{result['message']}")
            return

        if "result" in keys:
            result = result["result"]

        if isinstance(result, dict):
            keys = result.keys()
            if "capabilities" in keys:
                logger.debug(f"LSP Capabilities Response:\n{result}")

        if isinstance(result, list):
            ...

        if isinstance(result, tuple):
            ...